@@ -7,10 +7,15 @@ A Twisted wrapper for the asynchronous features of the PostgreSQL psycopg2
 driver.
 """
 
-import psycopg2
+try:
+    import psycopg2
+except ImportError:
+    # try psycopg2-ctypes
+    import psycopg2ct as psycopg2
+
 from zope.interface import implements
 
-from twisted.internet import interfaces, defer
+from twisted.internet import interfaces, main, defer
 from twisted.python import failure, log
 
 
@@ -30,6 +35,14 @@ class UnexpectedPollResult(Exception):
     """
 
 
+class AlreadyPolling(Exception):
+    """
+    The previous poll cycle has not been finished yet.
+
+    This probably indicates an issue in txpostgres, rather than in user code.
+    """
+
+
 class _CancelInProgress(Exception):
     """
     A query cancellation is in progress.
@@ -75,63 +88,106 @@ class _PollingMixin(object):
         @return: A Deferred that will fire with an instance of this class when
             the pollable reaches the OK state.
         """
-        if not self._pollingD:
-            self._pollingD = defer.Deferred(self._cancel)
-            # transform a psycopg2 QueryCanceledError into CancelledError
-            self._pollingD.addErrback(self._handleCancellation)
-        ret = self._pollingD
+        # this should never be called while the previous Deferred is still
+        # active, as it would clobber its reference
+        if self._pollingD:
+            return defer.fail(AlreadyPolling())
 
-        try:
-            self._pollingState = self.pollable().poll()
-        except:
-            d, self._pollingD = self._pollingD, None
-            d.errback()
-            return ret
+        ret = self._pollingD = defer.Deferred(self._cancel)
+        # transform a psycopg2 QueryCanceledError into CancelledError
+        self._pollingD.addErrback(self._handleCancellation)
 
-        if self._pollingState == psycopg2.extensions.POLL_OK:
-            d, self._pollingD = self._pollingD, None
-            d.callback(self)
-        elif self._pollingState == psycopg2.extensions.POLL_WRITE:
-            self.reactor.addWriter(self)
-        elif self._pollingState == psycopg2.extensions.POLL_READ:
-            self.reactor.addReader(self)
-        else:
-            d, self._pollingD = self._pollingD, None
-            d.errback(UnexpectedPollResult())
+        self.continuePolling()
 
         return ret
 
+    def continuePolling(self, swallowErrors=False):
+        """
+        Move forward in the poll cycle. This will call poll() on the wrapped
+        pollable and either wait for more I/O or callback or errback the
+        C{Deferred} returned earlier if the polling cycle has been
+        completed.
+
+        @type swallowErrors: C{bool}
+        @param swallowErrors: Should errors with no one to report them to be
+            ignored.
+        """
+        # This method often gets called from the reactor's doRead/doWrite
+        # handlers. Don't callback or errback the polling Deferred here, as
+        # arbitrary user code can be run by that and we don't want to deal with
+        # reentrancy issues if this user code tries running queries. The
+        # polling Deferred might also be simply not present, if we got called
+        # from a doRead after receiving a NOTIFY event.
+
+        try:
+            state = self.pollable().poll()
+        except:
+            if self._pollingD:
+                d, self._pollingD = self._pollingD, None
+                self.reactor.callLater(0, d.errback, failure.Failure())
+            elif not swallowErrors:
+                # no one to report the error to
+                raise
+        else:
+            if state == psycopg2.extensions.POLL_OK:
+                if self._pollingD:
+                    d, self._pollingD = self._pollingD, None
+                    self.reactor.callLater(0, d.callback, self)
+            elif state == psycopg2.extensions.POLL_WRITE:
+                self.reactor.addWriter(self)
+            elif state == psycopg2.extensions.POLL_READ:
+                self.reactor.addReader(self)
+            else:
+                if self._pollingD:
+                    d, self._pollingD = self._pollingD, None
+                    self.reactor.callLater(0, d.errback, UnexpectedPollResult())
+                elif not swallowErrors:
+                    # no one to report the error to
+                    raise UnexpectedPollResult()
+
     def doRead(self):
         self.reactor.removeReader(self)
-        self.poll()
+        if not self.pollable().closed:
+            self.continuePolling()
 
     def doWrite(self):
         self.reactor.removeWriter(self)
-        self.poll()
+        if not self.pollable().closed:
+            self.continuePolling()
 
     def logPrefix(self):
         return self.prefix
 
     def fileno(self):
-        if self.pollable():
-            return self.pollable().fileno()
-        else:
+        # this should never get called after the pollable has been
+        # disconnected, but Twisted versions affected by bug #4539 might cause
+        # it to happen, in which case we should return -1
+        if self.pollable().closed:
             return -1
 
+        return self.pollable().fileno()
+
     def connectionLost(self, reason):
-        # Do not errback self._pollingD here! We need to keep on calling poll()
-        # until it reports an error, which will errback self._pollingD with the
-        # correct failure. If we errback here, we won't finish the poll()
-        # cycle, which would leave psycopg2 in a state where it thinks there's
-        # still an async query underway.
+        # Do not errback self._pollingD here if the connection is still open!
+        # We need to keep on polling until it reports an error, which will
+        # errback self._pollingD with the correct failure. If we errback here,
+        # we won't finish the polling cycle, which would leave psycopg2 in a
+        # state where it thinks there's still an async query underway.
         #
         # If the connection got lost right after the first poll(), the Deferred
         # returned from it will never fire, leaving the caller hanging forever,
         # unless we push the connection state forward here. OTOH, if the
-        # connection is already closed, there's no pollable to poll, so check
-        # that before calling poll().
-        if self.pollable():
-            self.poll()
+        # connection is already closed, there's no pollable to poll, so if
+        # self._pollingD is still present, the only option is to errback it to
+        # prevent its waiters from hanging (you can't poll() a closed psycopg2
+        # connection)
+        if not self.pollable().closed:
+            # we're pushing the polling cycle to report pending failures, so if
+            # there's no one to report them to, swallow them
+            self.continuePolling(swallowErrors=True)
+        elif self._pollingD:
+            d, self._pollingD = self._pollingD, None
+            d.errback(reason)
 
     def _cancel(self, d):
         try:
@@ -148,6 +204,14 @@ class _PollingMixin(object):
         f.trap(psycopg2.extensions.QueryCanceledError)
         return failure.Failure(defer.CancelledError())
 
+    # Hack required to work with the Gtk2 reactor in Twisted <=11.0, which
+    # tries to access the "disconnected" property on the IReadWriteDescriptor
+    # it polls. To avoid attribute errors, forward that access to the "closed"
+    # property of the underlying connection.
+    def disconnected(self):
+        return self.pollable().closed
+    disconnected = property(disconnected)
+
     # forward all other access to the underlying connection
     def __getattr__(self, name):
         return getattr(self.pollable(), name)
@@ -207,11 +271,19 @@ class Cursor(_PollingMixin):
         except:
             return defer.fail()
 
-        return self.poll()
+        # tell the connection that a cursor is starting its poll cycle
+        self._connection.cursorRunning(self)
+
+        def finishedAndPassthrough(ret):
+            # tell the connection that the poll cycle has finished
+            self._connection.cursorFinished(self)
+            return ret
+
+        d = self.poll()
+        return d.addBoth(finishedAndPassthrough)
 
     def close(self):
-        _cursor, self._cursor = self._cursor, None
-        return _cursor.close()
+        return self._cursor.close()
 
     def __getattr__(self, name):
         # the pollable is the connection, but the wrapped object is the cursor
@@ -259,7 +331,7 @@ class Connection(_PollingMixin):
     @ivar cursorFactory: The factory used to produce cursors.
     """
 
-    connectionFactory = psycopg2.connect
+    connectionFactory = staticmethod(psycopg2.connect)
     cursorFactory = Cursor
 
     def __init__(self, reactor=None):
@@ -271,6 +343,10 @@ class Connection(_PollingMixin):
         # this lock will be used to prevent concurrent query execution
         self.lock = defer.DeferredLock()
         self._connection = None
+        # a set of cursors that should be notified about a disconnection
+        self._cursors = set()
+        # observers for NOTIFY events
+        self._notifyObservers = set()
 
     def pollable(self):
         return self._connection
@@ -285,7 +361,7 @@ class Connection(_PollingMixin):
         @rtype: C{Deferred}
         @returns: A Deferred that will fire when the connection is open.
         """
-        if self._connection:
+        if self._connection and not self._connection.closed:
             return defer.fail(AlreadyConnected())
 
         kwargs['async'] = True
@@ -294,14 +370,42 @@ class Connection(_PollingMixin):
         except:
             return defer.fail()
 
-        return self.poll()
+        def startReadingAndPassthrough(ret):
+            self.reactor.addReader(self)
+            return ret
+
+        # The connection is always a reader in the reactor, to receive NOTIFY
+        # events immediately when they're available.
+        d = self.poll()
+        return d.addCallback(startReadingAndPassthrough)
 
     def close(self):
         """
         Close the connection and disconnect from the database.
         """
-        _connection, self._connection = self._connection, None
-        _connection.close()
+        # We'll be closing the underlying socket so stop watching it.
+        self.reactor.removeReader(self)
+        self.reactor.removeWriter(self)
+
+        # make it safe to call Connection.close() multiple times, psycopg2
+        # treats this as an error but we don't
+        if not self._connection.closed:
+            self._connection.close()
+
+        # The above closed the connection socket from C code. Normally we would
+        # get connectionLost called on all readers and writers of that socket,
+        # but not if we're using the epoll reactor. According to the epoll(2)
+        # man page, closing a file descriptor causes it to be removed from all
+        # epoll sets automatically. In that case, the reactor might not have
+        # the chance to notice that the connection has been closed. To cover
+        # that, call connectionLost explicitly on the Connection and all
+        # outstanding Cursors. It's OK if connectionLost ends up being called
+        # twice, as the second call will not have any effects.
+
+        for cursor in set(self._cursors):
+            cursor.connectionLost(main.CONNECTION_DONE)
+
+        self.connectionLost(main.CONNECTION_DONE)
 
     def cursor(self):
         """
@@ -444,6 +548,103 @@ class Connection(_PollingMixin):
         except _CancelInProgress:
             pass
 
+    def cursorRunning(self, cursor):
+        """
+        Called automatically when a L{txpostgres.Cursor} created by this
+        L{txpostgres.Connection} starts polling after executing a query. User
+        code should never have to call this method.
+        """
+        # The cursor will now proceed to poll the psycopg2 connection, so stop
+        # polling it ourselves until it's done. Failure to do so would result
+        # in the connection "stealing" the POLL_OK result that appears after
+        # the query is completed and the Deferred returned from the cursor's
+        # poll() will never fire.
+        self.reactor.removeReader(self)
+        self._cursors.add(cursor)
+
+    def cursorFinished(self, cursor):
+        """
+        Called automatically when a L{txpostgres.Cursor} created by this
+        L{txpostgres.Connection} is done with polling after executing a
+        query. User code should never have to call this method.
+        """
+        self._cursors.remove(cursor)
+        # The cursor is done polling, resume watching the connection for NOTIFY
+        # events. Be careful to check the connection state, because it might
+        # have been closed while the cursor was polling and adding ourselves as
+        # a reader to a closed connection would be an error.
+        if not self._connection.closed:
+            self.reactor.addReader(self)
+            # While cursor was running, some notifies could have been
+            # delivered, so check for them.
+            self.checkForNotifies()
+
+    def doRead(self):
+        # call superclass to handle the pending read event on the socket
+        _PollingMixin.doRead(self)
+
+        # check for NOTIFY events
+        self.checkForNotifies()
+
+        # continue watching for NOTIFY events, but be careful to check the
+        # connection state in case one of the notify handler function caused a
+        # disconnection
+        if not self._connection.closed:
+            self.reactor.addReader(self)
+
+    def checkForNotifies(self):
+        """
+        Check if NOTIFY events have been received and if so, dispatch them to
+        the registered observers. This is done automatically, user code should
+        never need to call this method.
+        """
+        while self._connection.notifies:
+            notify = self._connection.notifies.pop()
+            # don't iterate over self._notifyObservers directly because the
+            # observer function might call removeNotifyObserver, thus modifying
+            # the set while it's being iterated
+            for observer in self.getNotifyObservers():
+                # don't allow exceptions from observers to propagate, as this
+                # method is called from doRead -- exceptions here would lead to
+                # a disconnect
+                try:
+                    observer(notify)
+                except:
+                    log.err()
+
+    def addNotifyObserver(self, observer):
+        """
+        Add an observer function that will get called whenever a NOTIFY event
+        is delivered to this connection. Any number of observers can be added
+        to a connection. Adding an observer that's already been added is
+        ignored.
+
+        @type observer: Any callable
+        @param observer: A callable whose first argument is a
+            L{psycopg2.extensions.Notify}.
+        """
+        self._notifyObservers.add(observer)
+
+    def removeNotifyObserver(self, observer):
+        """
+        Remove a previously added observer function. Removing an observer
+        that's never been added will be ignored.
+
+        @type observer: Any callable
+        @param observer: A callable that should no longer be called on NOTIFY
+            events.
+        """
+        self._notifyObservers.discard(observer)
+
+    def getNotifyObservers(self):
+        """
+        Get the currently registered notify observers.
+
+        @rtype: C{set}
+        @return: An set of callables that will get called on NOTIFY events.
+        """
+        return set(self._notifyObservers)
+
 
 class ConnectionPool(object):
     """
@@ -496,8 +697,11 @@ class ConnectionPool(object):
         @rtype: C{Deferred}
         @return: A C{Deferred} that fires when all connection have succeeded.
         """
-        d = defer.gatherResults([c.connect(*self.connargs, **self.connkw)
-                                 for c in self.connections])
+        # use DeferredList here, as gatherResults only got a consumeErrors
+        # keyword argument in Twisted 11.1.0
+        d = defer.DeferredList([c.connect(*self.connargs, **self.connkw)
+                                 for c in self.connections],
+                               fireOnOneErrback=True, consumeErrors=True)
         return d.addCallback(lambda _: self)
 
     def close(self):
